/* 
 * FreeSWITCH Modular Media Switching Software Library / Soft-Switch Application
 * Copyright (C) 2005/2006, Anthony Minessale II <anthmct@yahoo.com>
 *
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is FreeSWITCH Modular Media Switching Software Library / Soft-Switch Application
 *
 * The Initial Developer of the Original Code is
 * Anthony Minessale II <anthmct@yahoo.com>
 * Portions created by the Initial Developer are Copyright (C)
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s):
 *
 * mod_axroute.c -- Client module for axrouted
 *
 */
#include <errno.h>
#include <poll.h>

#include <switch.h>
#include <utils/utils.h>

#include "proto.h"


struct request {
	int type;
	int id;
	int version;

	int keylen;
	char key[MAX_KEY_SIZE];
};

struct response {
	int type;
	int id;
	int invalid;
	int status;
	int cost;
	switch_time_t duration;
	switch_time_t rtt;

	int gwlen;
	char gateway[MAX_GATEWAY_SIZE + 1];
};

enum {
	SFLAG_NONE    = 0,
	SFLAG_OFFLINE = (1 << 0),
};

struct server {
	void *next;
	int flags;

	char *host;
	unsigned short port;
	switch_sockaddr_t *addr;
};

static struct {
	int running;			/*!< thread running */
	int timeout;			/*!< request timeout */
	int retries;			/*!< number of connection attempts */
	int ping_interval;		/*!< */

	switch_memory_pool_t *pool;	/*!< global mempool */
	struct server *server;		/*!< list of server */
} globals;

static struct {
	unsigned long queries;

} statistics;

/*
 *
 */
static int decode_response( struct n_buffer *nbuf, struct response *resp )
{
	if( !nbuf || !resp )
		return -1;

	if( buffer_getnumleft(nbuf) < 1 )
		return -1;

	memset( resp, 0, sizeof(struct response) );

	/* read request type */
	resp->type = buffer_getu8( nbuf );

	if( resp->type == REQ_TYPE_PONG )
		return 0;

	if( buffer_getnumleft(nbuf) < 10 )
		return -1;

	while( buffer_getnumleft(nbuf) >= 2 ) {
		struct proto_type *pt;
		int8_t len = 0, type, id;

		id   = buffer_getu8( nbuf );
		type = buffer_getu8( nbuf );

		pt = lookup_type( id );
		if( !pt )					/* var ok? */
			return -1;

		if( pt->type != type )				/* type ok? */
			return -1;

		if( !(pt->flags & FLAGS_RESPONSE) )		/* variable allowed in response? */
			return -1;

		if( pt->len ) {					/* length needed, ok and enough bytes left? */
			len = buffer_getu8( nbuf );

			if( pt->len < len )
				return -1;

			if( buffer_getnumleft(nbuf) < len )
				return -1;
		}

		switch( id ) {
		case VAR_STATUS:
			resp->status = buffer_getu32( nbuf );
			break;

		case VAR_INVALID:
			resp->invalid = buffer_getu32( nbuf );
			break;

		case VAR_COST:
			resp->cost = buffer_getu32( nbuf );
			break;

		case VAR_GATEWAY:
			if( len > MAX_GATEWAY_SIZE )
				return -1;

			resp->gwlen = buffer_read( nbuf, resp->gateway, len );
			resp->gateway[resp->gwlen] = '\0';
			break;

		case VAR_DURATION:
			resp->duration = buffer_getu64( nbuf );
			break;

		default:
			return -1;
		}
	}
	return 0;
}

static int encode_request( struct n_buffer *nbuf, struct request *req )
{
	char vars[10];
	int n = 0;

	if( !nbuf || !req )
		return -1;

	memset( vars, 0, sizeof(vars) );

	/* request type */
	buffer_putu8( nbuf, req->type );

	if( req->type != REQ_TYPE_LOOKUP );
		return 0;

	/* status */
	vars[n++] = VAR_VERSION;
	vars[n++] = VAR_LOOKUP;

	/* request id */
	if( req->id )
		vars[n++] = VAR_REQUESTID;

	for( int x = 0; x < n; x++ ) {
		struct proto_type *pt;
		int8_t len = 0;

		pt = lookup_type( vars[x] );
		if( !pt )
			return -1;

		buffer_putu8( nbuf, pt->id );
		buffer_putu8( nbuf, pt->type );

		if( pt->id == VAR_LOOKUP )
			len = req->keylen;

		if( len ) {
			buffer_putu8( nbuf, len );
		}
		else if( pt->len ) {
			buffer_putu8( nbuf, pt->len );
		}

		switch( pt->id ) {
		case VAR_REQUESTID:
			buffer_putu32( nbuf, req->id );
			break;

		case VAR_VERSION:
			buffer_putu32( nbuf, req->version );
			break;

		case VAR_LOOKUP:
			buffer_write( nbuf, req->key, req->keylen );
			break;

		default:
			return -1;
		}
	}
	return 0;
}

static switch_status_t ping_server( struct server *srv, switch_memory_pool_t *pool )
{
	struct n_buffer *nbuf;
	struct request req;
	struct response resp;
	char buf[1024];
	switch_status_t res;
	switch_pollfd_t *pollfd;
	switch_socket_t *sock = NULL;
	switch_status_t status = SWITCH_STATUS_FALSE;
	switch_time_t before, after;
	switch_size_t len;
	int retry_cnt = 0;

	if( !srv || !pool )
		return SWITCH_STATUS_GENERR;

	memset( &req,  0, sizeof(req) );
	memset( &resp, 0, sizeof(resp) );

	res = switch_socket_create(&sock, PF_INET, SOCK_DGRAM, 0, pool);
	if( res != SWITCH_STATUS_SUCCESS ) {
		switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_DEBUG, "Error creating socket\n");
		goto done;
	}

	/*
	 * create socket
	 */
	do {
		/*
		 * Retry handling
		 */
		if( retry_cnt++ > globals.retries ) {
			goto done;
		}

		res = switch_socket_connect(sock, srv->addr);
		if( res != SWITCH_STATUS_SUCCESS ) {
			switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_DEBUG, "Error connecting socket\n");
			goto done;
		}

		nbuf = buffer_init( buf, sizeof(buf) );
		if( !nbuf ) {
			switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_DEBUG, "Failed to initialize buffer\n");
			goto done;
		}

		req.type = REQ_TYPE_PING;

		res = encode_request( nbuf, &req );
		if( res < 0 ) {
			switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_DEBUG, "Failed to encode request\n");
			goto done;
		}

		before = switch_time_now();

		len = buffer_getlen(nbuf);

		res = switch_socket_send(sock, buffer_getptr(nbuf), &len);
		if( res != SWITCH_STATUS_SUCCESS ) {
			switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_DEBUG, "Failed to send request: %s\n", strerror(errno));
			continue;
		}

		/*
		 * Receive response
		 */
		res = switch_socket_create_pollfd(&pollfd, sock, SWITCH_POLLIN, pool);
		if (res != SWITCH_STATUS_SUCCESS) {
			switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_DEBUG, "Failed to create pollfd for socket\n");
			goto done;
		}

		if( (res = switch_socket_waitfor(pollfd, globals.timeout)) == 1 ) {
			buffer_reset( nbuf );

			len = buffer_getsize(nbuf);

			res = switch_socket_recv(sock, buffer_getptr(nbuf), &len);
			if( res != SWITCH_STATUS_SUCCESS ) {
				switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_DEBUG, "Receive error: %s\n", strerror(errno));
				continue;
			}
			buffer_setlen( nbuf, len );

			after = switch_time_now();

			res = decode_response( nbuf, &resp );
			if( res < 0 ) {
				switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_ERROR, "Failed to decode response\n");
				goto done;
			}

			/* handle */
			if( resp.id != req.id ) {
				switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_ERROR, "Request and Response ID do not match\n");
				goto done;
			}

			if( resp.type == REQ_TYPE_PONG ) {
				status = SWITCH_STATUS_SUCCESS;
				goto done;
			}
			else {
				/* error */
				switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_ERROR, "Got invalid response from remote\n" );
				goto done;
			}
		}
		else if( res == 0 ) {
			switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_ERROR, "Request timed out (%d of %d retries)\n", retry_cnt, globals.retries );
		}

	} while( 1 );
done:
	if( sock ) {
		switch_socket_close(sock);
	}
	return status;
}

static switch_status_t do_request( char *value, struct response *resp, switch_memory_pool_t *pool )
{
	struct n_buffer *nbuf;
	struct request req;
	char buf[1024];
	switch_status_t res;
	switch_pollfd_t *pollfd;
	switch_socket_t *sock = NULL;
	switch_status_t status = SWITCH_STATUS_FALSE;
	switch_time_t before, after;
	switch_size_t len;
	struct server *cur_srv = globals.server;
	int retry_cnt = 0;

	memset( &req, 0, sizeof(req) );
	memset( resp, 0, sizeof(resp) );

	/*
	 * Find first server
	 */
	while( cur_srv && (cur_srv->flags & SFLAG_OFFLINE) ) {
		cur_srv = cur_srv->next;
	}
	if( !cur_srv ) {
		switch_log_printf( SWITCH_CHANNEL_LOG, SWITCH_LOG_ERROR, "No usable servers found\n" );
		return SWITCH_STATUS_GENERR;
	}

	res = switch_socket_create(&sock, PF_INET, SOCK_DGRAM, 0, pool);
	if( res != SWITCH_STATUS_SUCCESS ) {
		switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_DEBUG, "Error creating socket\n");
		goto done;
	}

	/*
	 * create socket
	 */
	do {
		res = switch_socket_connect(sock, cur_srv->addr);
		if( res != SWITCH_STATUS_SUCCESS ) {
			switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_DEBUG, "Error connecting socket\n");
			goto done;
		}

		nbuf = buffer_init( buf, sizeof(buf) );
		if( !nbuf ) {
			switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_DEBUG, "Failed to initialize buffer\n");
			goto done;
		}

		req.version = PROTOCOL_VERSION;
		req.keylen = strlen(value);
		memcpy(req.key, value, req.keylen);

		res = encode_request( nbuf, &req );
		if( res < 0 ) {
			switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_DEBUG, "Failed to encode request\n");
			goto done;
		}

		before = switch_time_now();

		len = buffer_getlen(nbuf);

		res = switch_socket_send(sock, buffer_getptr(nbuf), &len);
		if( res != SWITCH_STATUS_SUCCESS ) {
			switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_DEBUG, "Failed to send request: %s\n", strerror(errno));
			goto retry;
		}

		/*
		 * Receive response
		 */
		res = switch_socket_create_pollfd(&pollfd, sock, SWITCH_POLLIN, pool);
		if (res != SWITCH_STATUS_SUCCESS) {
			switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_DEBUG, "Failed to create pollfd for socket\n");
			goto done;
		}

		if( (res = switch_socket_waitfor(pollfd, globals.timeout)) == 1 ) {
			buffer_reset( nbuf );

			len = buffer_getsize(nbuf);

			res = switch_socket_recv(sock, buffer_getptr(nbuf), &len);
			if( res != SWITCH_STATUS_SUCCESS ) {
				switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_DEBUG, "Receive error: %s\n", strerror(errno));
				goto retry;
			}
			buffer_setlen( nbuf, len );

			after = switch_time_now();

			res = decode_response( nbuf, resp );
			if( res < 0 ) {
				switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_ERROR, "Failed to decode response\n");
				goto done;
			}

			/* handle */
			if( resp->id != req.id ) {
				switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_ERROR, "Request and Response ID do not match\n");
				goto done;
			}

			if( resp->status == 200 ) {
				/* ok */
				resp->gateway[resp->gwlen] = '\0';
				resp->rtt = after - before;

				status  = SWITCH_STATUS_SUCCESS;
				goto done;
			}
			else {
				/* error */
				switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_ERROR, "Got error response from remote: %d\n", resp->status);
				goto done;
			}
		}
		else if( res == 0 ) {
			switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_ERROR, "Request timed out (%d of %d retries)\n", retry_cnt, globals.retries );
		}

		/*
		 * Retry handling
		 */
retry:
		if( ++retry_cnt > globals.retries ) {
			/* set offline flag */
			cur_srv->flags |= SFLAG_OFFLINE;

			while( cur_srv && (cur_srv->flags & SFLAG_OFFLINE) ) {
				cur_srv = cur_srv->next;
			}

			if( !cur_srv ) {
				switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_ERROR, "Error after %d retries, request failed\n", retry_cnt );
				resp->status = 408;

				goto done;
			} else {
				switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_ERROR, "Error after %d retries, marking server offline and trying next one (%s:%hu)\n", retry_cnt, cur_srv->host, cur_srv->port );
				retry_cnt = 0;
			}
		}
	} while( 1 );
done:
	if( sock ) {
		switch_socket_close(sock);
	}
	return status;
}



SWITCH_MODULE_LOAD_FUNCTION(axroute_load);
SWITCH_MODULE_SHUTDOWN_FUNCTION(axroute_shutdown);
SWITCH_MODULE_RUNTIME_FUNCTION(axroute_runtime);
SWITCH_MODULE_DEFINITION(mod_axroute, axroute_load, axroute_shutdown, axroute_runtime);

SWITCH_STANDARD_APP(axroute_app_function)
{
	switch_channel_t *channel = switch_core_session_get_channel(session);
	switch_memory_pool_t *pool = switch_core_session_get_pool(session);
	char *dest = NULL, *appdata = NULL;
	char *argv[4] = { 0 };
	char buf[1024] = "";
	int argc = 0;

	if (!(appdata = switch_core_session_strdup(session, data))) {
		return;
	}

	if ((argc = switch_separate_string(appdata, ' ', argv, (sizeof(argv) / sizeof(argv[0]))))) {
		struct response resp;
		dest = argv[0];

		strcpy(buf, dest);
		if( do_request( buf, &resp, pool ) == 0 ) {
			switch_channel_set_variable(channel, "route_gateway", resp.gateway);

			sprintf(buf, "%" SWITCH_TIME_T_FMT, resp.rtt);
			switch_channel_set_variable(channel, "route_rtt_usecs", buf);

			if( resp.duration ) {
				sprintf(buf, "%" SWITCH_TIME_T_FMT, resp.duration);
				switch_channel_set_variable(channel, "route_query_nsecs", buf);
			}
		} else {
			switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_ERROR, "Route lookup failed\n");
		}

		sprintf(buf, "%d", (resp.status) ? resp.status : 500);
		switch_channel_set_variable(channel, "route_status", buf);

		statistics.queries++;
	}
}

SWITCH_STANDARD_API(axroute_function)
{
	char *dest = NULL, *appdata = NULL;
	char *argv[4] = { 0 };
//	char buf[1024] = "";
	int argc = 0;

	if (!(appdata = switch_safe_strdup(cmd))) {
		return SWITCH_STATUS_SUCCESS;
	}

	if ((argc = switch_separate_string(appdata, ' ', argv, (sizeof(argv) / sizeof(argv[0]))))) {
		struct response resp;
		dest = argv[0];

		if( do_request( dest, &resp, globals.pool ) == 0 ) {
			stream->write_function(stream, "Gateway   : %s\n", resp.gateway);
			stream->write_function(stream, "Query-RTT : %" SWITCH_TIME_T_FMT "\n", resp.rtt);

			if( resp.duration ) {
				stream->write_function(stream, "Query-Time: %" SWITCH_TIME_T_FMT "\n", resp.duration);
			}
			stream->write_function(stream, "Status    : %d\n", resp.status);
		} else {
			stream->write_function(stream, "Status: %d\n", resp.status);
			stream->write_function(stream, "-ERR lookup failed\n");
		}
		statistics.queries++;
	}

	if( appdata ) {
		switch_safe_free( appdata );
	}
	return SWITCH_STATUS_SUCCESS;
}

SWITCH_STANDARD_API(axroute_status_function)
{
	stream->write_function(stream, "Queries performed: %d", statistics.queries);

	return SWITCH_STATUS_SUCCESS;
}

SWITCH_MODULE_RUNTIME_FUNCTION(axroute_runtime)
{
	switch_status_t ret;
	struct server *srv = NULL;
	switch_time_t now = 0, last = 0;

	globals.running = 1;

	while( globals.running > 0 ) {

		now = switch_time_now();

		if( (now - last) > (globals.ping_interval * 1000) ) {
			for( srv = globals.server; srv; srv = srv->next ) {

				ret = ping_server( srv, globals.pool );
				if( ret != SWITCH_STATUS_SUCCESS ) {
					if( !(srv->flags & SFLAG_OFFLINE) ) {
						switch_log_printf( SWITCH_CHANNEL_LOG, SWITCH_LOG_ERROR, "Ping %s:%hu failed, marking offline\n", srv->host, srv->port );
						srv->flags |= SFLAG_OFFLINE;
					}
				} else {
					if( srv->flags & SFLAG_OFFLINE ) {
						switch_log_printf( SWITCH_CHANNEL_LOG, SWITCH_LOG_NOTICE, "Ping %s:%hu succeded, marking online\n", srv->host, srv->port );

						srv->flags &= ~SFLAG_OFFLINE;
					}
				}
			}
			last = switch_time_now();
		}
		switch_yield( 100000 );
	}

	globals.running = 0;

	return SWITCH_STATUS_SUCCESS;
}


static switch_status_t load_config(void)
{
	switch_xml_t cfg, xml = NULL, param, settings;
	switch_status_t status = SWITCH_STATUS_SUCCESS;
	struct server *server = NULL, *prev = NULL;
	char *cf = "axroute.conf";

	if (!(xml = switch_xml_open_cfg(cf, &cfg, NULL))) {
		switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_ERROR, "open of %s failed\n", cf);
		status = SWITCH_STATUS_FALSE;
		goto done;
	}

	if ((settings = switch_xml_child(cfg, "settings"))) {
		for (param = switch_xml_child(settings, "param"); param; param = param->next) {
			char *var = (char *) switch_xml_attr_soft(param, "name");
			char *val = (char *) switch_xml_attr_soft(param, "value");
			if (!strcasecmp(var, "query-timeout")) {
				globals.timeout = atoi(val);
			} else if (!strcasecmp(var, "query-retries")) {
				globals.retries = atoi(val);
			}
		}
	}

	if ((settings = switch_xml_child(cfg, "servers"))) {
		for (param = switch_xml_child(settings, "server"); param; param = param->next) {
			char *host = (char *) switch_xml_attr_soft(param, "host");
			char *port = (char *) switch_xml_attr_soft(param, "port");

			if (!host)
				continue;

			server = switch_core_alloc(globals.pool, sizeof(struct server));
			if( !server ) {
				status = SWITCH_STATUS_FALSE;
				goto done;
			}

			server->host  = switch_core_strdup(globals.pool, host);
			server->port  = atoi(port);
			server->flags = SFLAG_NONE;
			server->next  = NULL;

			if( !server->port ) {
				server->port = 1234;
			}

			if(switch_sockaddr_info_get(&server->addr, host, AF_INET, server->port, 0, globals.pool) != SWITCH_STATUS_SUCCESS) {
				switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_ERROR, "Failed to lookup server address\n");
				status = SWITCH_STATUS_FALSE;
				goto done;
			}

			if( prev ) {
				prev->next = server;
				prev       = server;
			} else {
				globals.server = server;
				prev           = server;
			}
		}
	}

	if (!globals.server) {
		switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_ERROR, "No servers specified\n");
		status = SWITCH_STATUS_FALSE;
		goto done;		
	}

	if (!globals.timeout) {
		globals.timeout = 3000;
	}

	if (!globals.retries) {
		globals.retries = 3;
	}

	if (!globals.ping_interval) {
		globals.ping_interval = 60000;	/* 60 seconds */
	}

done:
	if (xml) {
		switch_xml_free(xml);
	}

	return status;
}

SWITCH_MODULE_LOAD_FUNCTION(axroute_load)
{
	switch_application_interface_t *app_interface;
	switch_api_interface_t *api_interface;

	memset(&statistics, 0, sizeof(statistics));
	memset(&globals, 0, sizeof(globals));

	switch_core_new_memory_pool(&globals.pool);

	if (load_config() != SWITCH_STATUS_SUCCESS) {
		switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_ERROR, "Failed to load configuration\n");
		return SWITCH_STATUS_GENERR;
	}

	/* connect my internal structure to the blank pointer passed to me */
	*module_interface = switch_loadable_module_create_module_interface(pool, modname);

	SWITCH_ADD_APP(app_interface, "axroute", "Perform an AXROUTE lookup", "Perform an AXROUTE lookup", axroute_app_function, "<number>",
				   SAF_SUPPORT_NOMEDIA);

	SWITCH_ADD_API(api_interface, "axroute", "Perform an AXROUTE lookup", axroute_function, "<number>");
	SWITCH_ADD_API(api_interface, "axroute_status", "axroute status", axroute_status_function, "");

	switch_log_printf(SWITCH_CHANNEL_LOG, SWITCH_LOG_NOTICE, "Hello World!\n");

	/* indicate that the module should continue to be loaded */
	return SWITCH_STATUS_NOUNLOAD;
}

SWITCH_MODULE_SHUTDOWN_FUNCTION(axroute_shutdown)
{
	int counter = 0;

	globals.running = -1;

	while( globals.running != 0 && counter++ < 1000 ) {
		switch_yield( 1000 );
	}

	return SWITCH_STATUS_SUCCESS;
}


/* For Emacs:
 * Local Variables:
 * mode:c
 * indent-tabs-mode:t
 * tab-width:4
 * c-basic-offset:4
 * End:
 * For VIM:
 * vim:set softtabstop=4 shiftwidth=4 tabstop=4 expandtab:
 */
