#ifndef __ATOMIC_H__
#define __ATOMIC_H__


/*
 * Memory fencing operations
 */
#if defined(__sparc__)
/*
 * sparc memory fencing instructions
 */
#define atomic_mb()  __asm__ volatile ( "membar #LoadLoad | #LoadStore | #StoreStore | #StoreLoad" ::: "memory" )
#define atomic_rmb() __asm__ volatile ( "membar #LoadLoad" ::: "memory" )
#define atomic_smb() __asm__ volatile ( "membar #StoreStore" ::: "memory" )

#elif defined(__x86_64__) || defined(__i686__)
/*
 * x86_64 memory fencing instructions
 */
#define atomic_mb()  __asm__ volatile ( ".byte 0x0f, 0xae, 0xf0" ::: "memory" )	// mfence
#define atomic_rmb() __asm__ volatile ( ".byte 0x0f, 0xae, 0xe8" ::: "memory" )	// lfence
#define atomic_smb() __asm__ volatile ( ".byte 0x0f, 0xae, 0xf8" ::: "memory" )	// sfence

#elif defined(__i686__)
/*
 * i686 memory fencing instructions
 */
#define atomic_mb()  __asm__ volatile ( ".byte 0x0f, 0xae, 0xf0" ::: "memory" )	// mfence
#define atomic_rmb() __asm__ volatile ( ".byte 0x0f, 0xae, 0x8e" ::: "memory" )	// lfence
#define atomic_smb() __asm__ volatile ( ".byte 0x0f, 0xae, 0xf8" ::: "memory" )	// sfence

#else
/*
 * generic memory fencing instructions
 */
#define atomic_mb()  __asm__ volatile ( "" ::: "memory" )
#define atomic_rmb() __asm__ volatile ( "" ::: "memory" )
#define atomic_smb() __asm__ volatile ( "" ::: "memory" )

#endif



/*
 * Atomic operations
 */
#if defined(__i686__) || defined(__x86_64__)

typedef struct { volatile int value; } atomic_t;

#if defined(__x86_64__)

#define atomic_set32(ptr, val)		((ptr)->value = val)
#define atomic_set32_fenced(ptr, val)	({ (ptr)->value = val; atomic_smb(); })
#define atomic_read32(ptr)		((ptr)->value)
#define atomic_read32_fenced(ptr)	({ atomic_rmb(); (ptr)->value; })

#define atomic_inc32(ptr) __asm__ volatile ( "lock; incl %0" : "=m"((ptr)->value) : "m"((ptr)->value) )
#define atomic_dec32(ptr) __asm__ volatile ( "lock; decl %0" : "=m"((ptr)->value) : "m"((ptr)->value) )

#define atomic_inc32_and_test(ptr) ({ 				\
		int ret;					\
		__asm__ volatile ( "lock; incl %0; sete %1"	\
				: "=m"((ptr)->value), "=q"(ret)	\
				: "m"((ptr)->value)		\
				: "memory" );			\
		ret;						\
	})

#define atomic_dec32_and_test(ptr) ({ 				\
		int ret;					\
		__asm__ volatile ( "lock; decl %0; sete %1"	\
				: "=m"((ptr)->value), "=q"(ret)	\
				: "m"((ptr)->value)		\
				: "memory" );			\
		ret;						\
	})

#define atomic_add32(ptr, x) __asm__ volatile ( "lock; addl %1,%0" : "=m"((ptr)->value) : "ir"(x), "m"((ptr)->value) )
#define atomic_sub32(ptr, x) __asm__ volatile ( "lock; subl %1,%0" : "=m"((ptr)->value) : "ir"(x), "m"((ptr)->value) )

#define atomic_add32_and_test(ptr, x) ({ 			\
		int ret;					\
		__asm__ volatile ( "lock; addl %2,%0; sete %1"	\
				: "=m"((ptr)->value), "=q"(ret)	\
				: "ir"(x), "m"((ptr)->value)	\
				: "memory" );			\
		ret;						\
	})

#define atomic_sub32_and_test(ptr, x) ({ 			\
		int ret;					\
		__asm__ volatile ( "lock; subl %2,%0; sete %1"	\
				: "=m"((ptr)->value), "=q"(ret)	\
				: "ir"(x), "m"((ptr)->value)	\
				: "memory" );			\
		ret;						\
	})

#define atomic_cmpxchg32(ptr, old, new)	({				\
		int ret;						\
		__asm__ volatile ( "lock; cmpxchgl %1, %2"		\
				: "=a"(ret)				\
				: "q"(new), "m"((ptr)->value), "0"(old)	\
				: "memory");				\
		ret;							\
	})

/* 64bit atomic ops */
#define atomic_set64(ptr, val)		((ptr)->value = val)
#define atomic_set64_fenced(ptr, val)	({ (ptr)->value = val; atomic_smb(); })
#define atomic_read64(ptr)		((ptr)->value)
#define atomic_read64_fenced(ptr)	({ atomic_rmb(); (ptr)->value; })

#define atomic_inc64(ptr) __asm__ volatile ( "lock; incq %0" : "=m"((ptr)->value) : "m"((ptr)->value) )
#define atomic_dec64(ptr) __asm__ volatile ( "lock; decq %0" : "=m"((ptr)->value) : "m"((ptr)->value) )

#define atomic_inc64_and_test(ptr) ({ 				\
		int ret;					\
		__asm__ volatile ( "lock; incq %0; sete %1"	\
				: "=m"((ptr)->value), "=q"(ret)	\
				: "m"((ptr)->value)		\
				: "memory" );			\
		ret;						\
	})

#define atomic_dec64_and_test(ptr) ({ 				\
		int ret;					\
		__asm__ volatile ( "lock; decq %0; sete %1"	\
				: "=m"((ptr)->value), "=q"(ret)	\
				: "m"((ptr)->value)		\
				: "memory" );			\
		ret;						\
	})

#define atomic_add64(ptr, x) __asm__ volatile ( "lock; addq %1,%0" : "=m"((ptr)->value) : "ir"(x), "m"((ptr)->value) )
#define atomic_sub64(ptr, x) __asm__ volatile ( "lock; subq %1,%0" : "=m"((ptr)->value) : "ir"(x), "m"((ptr)->value) )

#define atomic_add64_and_test(ptr, x) ({ 			\
		int ret;					\
		__asm__ volatile ( "lock; addq %2,%0; sete %1"	\
				: "=m"((ptr)->value), "=q"(ret)	\
				: "ir"(x), "m"((ptr)->value)	\
				: "memory" );			\
		ret;						\
	})

#define atomic_sub64_and_test(ptr, x) ({ 			\
		int ret;					\
		__asm__ volatile ( "lock; subq %2,%0; sete %1"	\
				: "=m"((ptr)->value), "=q"(ret)	\
				: "ir"(x), "m"((ptr)->value)	\
				: "memory" );			\
		ret;						\
	})


#define atomic_cmpxchg64(ptr, old, new)	({				\
		int ret;						\
		__asm__ volatile ( "lock; cmpxchgq %1, %2"		\
				: "=a"(ret)				\
				: "q"(new), "m"((ptr)->value), "0"(old)	\
				: "memory");				\
		ret;							\
	})


#elif defined(__i686__)
/*
 * x86 atomic instructions
 */

#define atomic_set32(ptr, val)		((ptr)->value = val)
#define atomic_set32_fenced(ptr, val)	({ (ptr)->value = val; atomic_smb(); })
#define atomic_read32(ptr)		((ptr)->value)
#define atomic_read32_fenced(ptr)	({ atomic_rmb(); (ptr)->value; })

#define atomic_inc32(ptr) __asm__ volatile ( "lock; incl %0" : "=m"((ptr)->value) : "m"((ptr)->value) )
#define atomic_dec32(ptr) __asm__ volatile ( "lock; decl %0" : "=m"((ptr)->value) : "m"((ptr)->value) )

#define atomic_inc32_and_test(ptr) ({ 				\
		int ret;					\
		__asm__ volatile ( "lock; incl %0; sete %1"	\
				: "=m"((ptr)->value), "=q"(ret)	\
				: "m"((ptr)->value)		\
				: "memory" );			\
		ret;						\
	})

#define atomic_dec32_and_test(ptr) ({ 				\
		int ret;					\
		__asm__ volatile ( "lock; decl %0; sete %1"	\
				: "=m"((ptr)->value), "=q"(ret)	\
				: "m"((ptr)->value)		\
				: "memory" );			\
		ret;						\
	})

#define atomic_add32(ptr, x) __asm__ volatile ( "lock; addl %1,%0" : "=m"((ptr)->value) : "ir"(x), "m"((ptr)->value) )
#define atomic_sub32(ptr, x) __asm__ volatile ( "lock; subl %1,%0" : "=m"((ptr)->value) : "ir"(x), "m"((ptr)->value) )

#define atomic_add32_and_test(ptr, x) ({ 			\
		int ret;					\
		__asm__ volatile ( "lock; addl %2,%0; sete %1"	\
				: "=m"((ptr)->value), "=q"(ret)	\
				: "ir"(x), "m"((ptr)->value)	\
				: "memory" );			\
		ret;						\
	})

#define atomic_sub32_and_test(ptr, x) ({ 			\
		int ret;					\
		__asm__ volatile ( "lock; subl %2,%0; sete %1"	\
				: "=m"((ptr)->value), "=q"(ret)	\
				: "ir"(x), "m"((ptr)->value)	\
				: "memory" );			\
		ret;						\
	})

#define atomic_cmpxchg32(ptr, old, new)	({				\
		int ret;						\
		__asm__ volatile ( "lock; cmpxchgl %1, %2"		\
				: "=a"(ret)				\
				: "q"(new), "m"((ptr)->value), "0"(old)	\
				: "memory" );				\
		ret;							\
	})

#endif

#else	/* no native atomic ops support */
#include <pthread.h>

/*
 * Generic atomic support
 */
typedef struct {
	pthread_spinlock_t lock;
	int value;
} atomic_t;

#define atomic_set32(ptr, val)	({						\
		pthread_spin_init(&(ptr)->lock, PTHREAD_PROCESS_SHARED);	\
		(ptr)->value = val;						\
	})

#define atomic_set32_fenced(ptr, val)	({					\
		pthread_spin_init(&(ptr)->lock, PTHREAD_PROCESS_SHARED);	\
		(ptr)->value = val;						\
		atomic_smb();							\
	})

#define atomic_read32(ptr)		((ptr)->value)
#define atomic_read32_fenced(ptr)	({ atomic_rmb(); (ptr)->value; })

#define atomic_inc32(ptr)	({			\
		pthread_spin_lock(&(ptr)->lock);	\
		(ptr)->value++;				\
		pthread_spin_unlock(&(ptr)->lock);	\
	})

#define atomic_dec32(ptr)	({			\
		pthread_spin_lock(&(ptr)->lock);	\
		(ptr)->value--;				\
		pthread_spin_unlock(&(ptr)->lock);	\
	})

#define atomic_inc32_and_test(ptr, n)	({		\
		int ret;				\
		pthread_spin_lock(&(ptr)->lock);	\
		(ptr)->value += n;			\
		ret = ((ptr)->value == 0);		\
		pthread_spin_unlock(&(ptr)->lock);	\
		ret;					\
	})

#define atomic_dec32_and_test(ptr, n)	({		\
		int ret;				\
		pthread_spin_lock(&(ptr)->lock);	\
		(ptr)->value -= n;			\
		ret = ((ptr)->value == 0);		\
		pthread_spin_unlock(&(ptr)->lock);	\
		ret;					\
	})

#define atomic_cmpxchg32(ptr, old, new)	({		\
		int ret;				\
		pthread_spin_lock(&(ptr)->lock);	\
		ret = (ptr)->value;			\
		if((ptr)->value == old)			\
			(ptr)->value = new;		\
		pthread_spin_unlock(&(ptr)->lock);	\
		ret;					\
	})

#endif

inline int atomic_fetch_and_add32(atomic_t *ptr, const int n)
{
	int prev;

	do {
		prev = atomic_read32(ptr);
	}
	while(atomic_cmpxchg32(ptr, prev, prev + n) != prev);

	return prev;
}

inline int atomic_fetch_and_sub32(atomic_t *ptr, const int n)
{
	int prev;

	do {
		prev = atomic_read32(ptr);
	}
	while(atomic_cmpxchg32(ptr, prev, prev - n) != prev);

	return prev;
}

#endif
