
#include "common.h"
#include "bitmap.h"

//#define BITMAP_DEBUG

int bitmap_create( unsigned long **bm, const ssize_t size )
{
	ssize_t bytes;

	/* round size to multiple of nbytes */
	bytes = BITMAP_SIZE(size);

	*bm = malloc( bytes );
	if( !*bm ) {
		errno = ENOSPC;
		return -1;
	}

	memset( *bm, 0, bytes );

#ifdef BITMAP_DEBUG
	fprintf( stderr, "\t++ Creating bitmap with size %d (%d bytes)\n", size, bytes );
#endif

	return 0;
}

typedef enum {
	BITMAP_CLEAR = 0,
	BITMAP_SET,
	BITMAP_CHECK
} bitmap_op;


#define BITMAP_ROW(pos) \
	( ( pos ) / BITS_PER_LONG )

#define BITMAP_MASK(pos) \
	1 << ( ( pos ) % BITS_PER_LONG )


static inline int __bitmap_op( unsigned long *bm, const int nbits, bitmap_op op, int pos )
{
	int	row;
	long	mask;

	if( pos < 0 || pos >= nbits ) {
#ifdef BITMAP_DEBUG
		fprintf( stderr, "\t-- Bitmap operation out of bounds: %d vs. %d\n", pos, nbits );
#endif
		errno = EINVAL;
		return -1;
	}

	row  = BITMAP_ROW(pos);
	mask = BITMAP_MASK(pos);

	switch(op) {
	case BITMAP_CLEAR:
#ifdef BITMAP_DEBUG
		fprintf( stderr, "\t++ Clearing bit for %d @ row %d (mask %#lx -> %#lx)\n", pos, row, mask, ~mask );
#endif
		bm[row] &= ~mask;
		break;

	case BITMAP_SET:
#ifdef BITMAP_DEBUG
		fprintf( stderr, "\t++ Setting bit for %d @ row %d (mask %#lx)\n", pos, row, mask );
#endif
		bm[row] |= mask;
		break;

	case BITMAP_CHECK:
		return ( bm[row] & mask ) ? 1 : 0;
	}

	return 0;
}

int bitmap_is_set( unsigned long *bm, const int nbits, const int pos )
{
	return __bitmap_op( bm, nbits, BITMAP_CHECK, pos );
}

int bitmap_set( unsigned long *bm, const int nbits, const int pos )
{
	return __bitmap_op( bm, nbits, BITMAP_SET, pos );
}

int bitmap_clear( unsigned long *bm, const int nbits, const int pos )
{
	return __bitmap_op( bm, nbits, BITMAP_CLEAR, pos );
}

void bitmap_clear_all( unsigned long *bm, const int nbits )
{
	memset( bm, 0, BITMAP_SIZE(nbits) );
}

int bitmap_is_empty( const unsigned long *bm, const int nbits ) {
	int i;
	unsigned long *p = (unsigned long *)bm;

	for( i=0; i < BITMAP_SIZE(nbits); i++ ) {
		if( *(p++) )
			return 0;
	}
	return 1;
}

int bitmap_first_bit( const unsigned long *bm, const int nbits ) {
	unsigned long mask = 1;
	unsigned long *ptr = (unsigned long *)bm;
	int i = 1;

	while( i <= nbits ) {
		if( *ptr & mask )
			return i;

		if( mask & BITMAP_TOP_BIT ) {
			ptr++;
			mask = 1;
		}
		else
			mask <<= 1;

		i++;
	}

	return 0;
}
