#ifndef _BUFFER_H
#define _BUFFER_H

/*
 * Buffer
 */

/**
 * Network buffer
 */
struct n_buffer {
	long size;	/*!< Size of the buffer (in bytes) */
	long end_pos;	/*!< End pointer, marking the last data byte in the buffer */
	long cur_pos;	/*!< Current read/write position */

	char raw[0];	/*!< Raw data */
};

/**
 * buffer_init
 * \brief	Initialize statically allocated buffer
 * \param[in]	buf	Buffer
 * \param[in]	size	Size of buffer
 * \return	initialized buffer, NULL on error
 */
struct n_buffer *buffer_init( char *buf, int size );

/**
 * buffer_create
 * \brief	Create a new network buffer
 * \param[in]	size	Size of the new buffer
 * \return	the new buffer, NULL on error
 */
struct n_buffer *buffer_create( int size );


/**
 * buffer_destroy
 * \brief	Destroy a network buffer
 * \param[in]	buf	network buffer
 */
void buffer_destroy( struct n_buffer *buf );


/**
 * buffer_reset
 * \brief	Reset buffer, zeroing the content
 * \param[in]	buf	network buffer
 */
void buffer_reset( struct n_buffer *buf );


/**
 * buffer_getpos
 * \brief	Get (absolute) position of read/write pointer
 * \param[in]	buf	network buffer
 * \return	current position
 */
long buffer_getpos( const struct n_buffer *buf );

/**
 * buffer_setpos
 * \brief	Set (absolute) position of read/write pointer
 * \param[in]	buf	network buffer
 * \param[in]	pos	new position
 */
void buffer_setpos( struct n_buffer *buf, const long pos );


/**
 * Direction / starting point of seek operation
 */
typedef enum {
	BUFFER_SEEK_SET = 0,	/*!< Seek from the beginning of the buffer */
	BUFFER_SEEK_CUR,	/*!< Seek from the current position */
	BUFFER_SEEK_END		/*!< Seek from the end of the buffer */
} buffer_seek_t;

/**
 * buffer_seek
 * \brief	Reposition the buffer's read/write pointer
 * \param[in]	buf	network buffer
 * \param[in]	type	starting point of seek
 * \param[in]	pos	seek position / offset
 *
 * \author	Stefan Knoblich <s.knoblich@axsentis.de>
 */
void buffer_seek( struct n_buffer *buf, const buffer_seek_t type, const long pos );



/*
 * Low-level direct content access
 */

/**
 * buffer_getsize
 * \brief	Size of the buffer
 * \param[in]	buf	network buffer
 * \return	size in bytes
 */
long buffer_getsize( const struct n_buffer *buf );


/**
 * buffer_getptr
 * \brief	Return a pointer to the the internal data storage
 * \param[in]	buf	network buffer
 * \return	pointer to the data storage
 */
char * buffer_getptr( const struct n_buffer *buf );


/**
 * buffer_getcurrent
 * \brief	Return a pointer to the current position in the internal data storage
 * \param[in]	buf	network buffer
 * \return	pointer to the data storage
 */
char * buffer_getcurrent( const struct n_buffer *buf );


/**
 * buffer_getlen
 * \brief	Get the position of the end pointer (= number of data bytes in the buffer)
 * \param[in]	buf	network buffer
 * \return	number of bytes in the buffer
 *
 * \bug		Description is wrong(?)
 */
long buffer_getlen( const struct n_buffer *buf );


/**
 * buffer_setlen
 * \brief	Set the end pointer to a new (absolute) position
 * \param[in]	buf	network buffer
 * \param[in]	len	new position of the end pointer (= amount of data in the buffer)
 *
 * \bug		Description is wrong(?)
 */
void buffer_setlen( struct n_buffer *buf, const long len );


/**
 * buffer_getnumleft
 * \brief	Get the number of bytes left to read
 * \param[in]	buf	network buffer
 * \return	number of bytes left to read
 */
long buffer_getnumleft( const struct n_buffer *buf );


/**
 * buffer_getnumfree
 * \brief	Get the number of bytes left to write
 * \param[in]	buf	network buffer
 * \return	number of bytes of free space left
 */
long buffer_getnumfree( const struct n_buffer *buf );


/*
 * Higher-level access
 */

/**
 * buffer_read
 * \brief	Read at max. n bytes from the network buffer
 * \param[in]	buf	network buffer
 * \param[out]	ptr	copy data to this location
 * \param[in]	num	number of bytes to read
 * \return	number of bytes read or -1 on error
 *
 * \author	Stefan Knoblich <s.knoblich@axsentis.de>
 */
size_t buffer_read( struct n_buffer *buf, void *ptr, size_t num );


/**
 * buffer_write
 * \brief	Write at max. n bytes to the network buffer
 * \note	Updates the end pointer
 * \param[out]	buf	network buffer
 * \param[in]	ptr	copy data from this location
 * \param[in]	num	number of bytes to write
 * \return	number of bytes written or -1 on error
 *
 * \author	Stefan Knoblich <s.knoblich@axsentis.de>
 */
size_t buffer_write( struct n_buffer *buf, const void *ptr, size_t num );


/**
 * buffer_get
 * \brief	Get n bytes from the network buffer
 * \param[in]	buf	network buffer
 * \param[in]	ptr	pointer to memory location to copy to
 * \param[in]	num	number of bytes to read
 * \return	0 on success, -1 on error (e.g. not enough data left)
 *
 */
int buffer_get( struct n_buffer *buf, void *ptr, size_t num );


/**
 * buffer_put
 * \brief	Put n bytes into the network buffer
 * \note	Updates the end pointer
 * \param[in]	buf	network buffer
 * \param[in]	ptr	pointer to memory location to copy from
 * \param[in]	num	number of bytes to write
 * \return	0 on success, -1 on error (e.g. not enough free space left)
 *
 */
int buffer_put( struct n_buffer *buf, const void *ptr, size_t num );


/**
 * buffer_memcmp
 * \brief	Compare n bytes between a network buffer and a memory location
 * \param[in]	buf	network buffer
 * \param[in]	val	memory location
 * \param[in]	size	number of bytes to compare
 * \return	0 if both match, -1 or 1 else
 *
 * \author	Stefan Knoblich <s.knoblich@axsentis.de>
 */
int buffer_memcmp( struct n_buffer *buf, const void *val, const size_t size );


/**
 * Get a scalar value from the network buffer
 */
#define buffer_getscalar(buf, type) ({		\
	type __tmp;				\
	buffer_get(buf, &__tmp, sizeof(type));	\
	( !__tmp ) ? 0 : __tmp; })

/**
 * Get a value from the network buffer (generic)
 */
#define buffer_getval(buf, var)	\
	buffer_get(buf, &var, sizeof(var))

/**
 * Put a scalar value into the network buffer
 */
#define buffer_putscalar(buf, var, type) ({	\
	const type _x_ = (type)var;		\
	buffer_put(buf, &_x_, sizeof(type));	\
	})


/**
 * Get a signed 1 byte value
 */
#define buffer_gets8(buf) \
	buffer_getscalar( buf, int8_t ) 

/**
 * Get an unsigned 1 byte value
 */
#define buffer_getu8(buf) \
	buffer_getscalar( buf, uint8_t )


/**
 * Get a signed 2 byte value
 */
#define buffer_gets16(buf) \
	buffer_getscalar( buf, int16_t )

/**
 * Get an unsigned 2 byte value
 */
#define buffer_getu16(buf) \
	buffer_getscalar( buf, uint16_t )


/**
 * Get a signed 4 byte value
 */
#define buffer_gets32(buf) \
	buffer_getscalar( buf, int32_t ) 

/**
 * Get an unsigned 4 byte value
 */
#define buffer_getu32(buf) \
	buffer_getscalar( buf, uint32_t )


/**
 * Get a signed 8 byte value
 */
#define buffer_gets64(buf) \
	buffer_getscalar( buf, int64_t ) 

/**
 * Get an unsigned 8 byte value
 */
#define buffer_getu64(buf) \
	buffer_getscalar( buf, uint64_t )



/**
 * Store a signed 1 byte value
 */
#define buffer_puts8(buf, x) \
	buffer_putscalar( buf, (x), int8_t ) 

/**
 * Store an unsigned 1 byte value
 */
#define buffer_putu8(buf, x) \
	buffer_putscalar( buf, (x), uint8_t )


/**
 * Store a signed 2 byte value
 */
#define buffer_puts16(buf, x) \
	buffer_putscalar( buf, (x), int16_t )

/**
 * Store an unsigned 2 byte value
 */
#define buffer_putu16(buf, x) \
	buffer_putscalar( buf, (x), uint16_t )


/**
 * Store a signed 4 byte value
 */
#define buffer_puts32(buf, x) \
	buffer_putscalar( buf, (x), int32_t ) 

/**
 * Store an unsigned 4 byte value
 */
#define buffer_putu32(buf, x) \
	buffer_putscalar( buf, (x), uint32_t )

/**
 * Store a signed 8 byte value
 */
#define buffer_puts64(buf, x) \
	buffer_putscalar( buf, (x), int64_t ) 

/**
 * Store an unsigned 8 byte value
 */
#define buffer_putu64(buf, x) \
	buffer_putscalar( buf, (x), uint64_t )


/**
 * buffer_print
 * \brief	Print buffer content to screen (in hex)
 * \param[in]	buf	network buffer to print
 *
 * \author	Stefan Knoblich <s.knoblich@axsentis.de>
 */
void buffer_print( struct n_buffer *buf );

#endif
