
#include <sys/types.h>
#include <unistd.h>

#include "common.h"
#include "dlist.h"
#include "coroutine.h"


int coro_init( struct coro_scheduler *sched )
{
	sched->last_id      = 0;
	sched->abort        = 0;
	sched->active_tasks = 0;

	dlist_init_head( &sched->tasks );
	return 0;
}

#ifndef USE_SWAPCONTEXT
volatile static int coro_sig_handled;

// helper
static void coro_helper( int signum, siginfo_t *info, void *data )
{
	struct coro_task *tsk = info->si_ptr;

	if( !sigsetjmp( tsk->env, 0 ) ) {
		coro_sig_handled = 1;
		return;
	}

	tsk->func( tsk );

	siglongjmp( tsk->sched->env, 0 );
}
#endif

int coro_add_task( struct coro_scheduler *sched, void (* func)( struct coro_task *tsk ) )
{
	struct coro_task *new;
#ifndef USE_SWAPCONTEXT
	struct sigaction act, oldact;
	union sigval sval;
	stack_t ss;
#endif

	new = malloc( sizeof(struct coro_task) );
	if( !new )
		return -1;

	memset( new, 0, sizeof(struct coro_task) );

	new->id    = ++sched->last_id;
	new->func  = func;
	new->state = CORO_INIT;
	new->sched = sched;

#ifdef USE_SWAPCONTEXT
	new->env.uc_stack.ss_sp   = new->stack;
	new->env.uc_stack.ss_size = sizeof(new->stack);

	new->env.uc_link = &sched->env;
	getcontext( &new->env );
	makecontext( &new->env, (void (*)(void))new->func, 1, new );

#else
	coro_sig_handled = 0;

	ss.ss_sp    = new->stack;
	ss.ss_size  = sizeof(new->stack);
	ss.ss_flags = 0;

	if( sigaltstack( &ss, NULL ) < 0 ) {
		free( new );
		return -1;
	}

	/*
	 * Register new handler for SIGUSR1
	 */
	sigfillset( &act.sa_mask );
	sigdelset( &act.sa_mask, SIGUSR1 );

	act.sa_sigaction = coro_helper;
	act.sa_flags = SA_ONESHOT | SA_SIGINFO | SA_ONSTACK;

	if( sigaction( SIGUSR1, &act, &oldact ) < 0 ) {
		free( new );
		return -1;
	}

	/*
	 * Trigger helper
	 */
	sval.sival_ptr = new;
	sigqueue( getpid(), SIGUSR1, sval );

	while( !coro_sig_handled );

	/*
	 * Disable custom stack and signal handler
	 */
	sigaltstack( NULL, &ss );
	ss.ss_flags = SS_DISABLE;

	if( sigaltstack( &ss, NULL ) < 0 )
		printf( "Failed to disable stack\n" );

	sigaction( SIGUSR1, &oldact, NULL );
#endif

	dlist_insert_tail( &sched->tasks, &new->list );

	return 0;
}

inline int coro_getid( const struct coro_task *tsk )
{
	return tsk->id;
}

int coro_main( struct coro_scheduler *sched )
{
	struct coro_task	*cur	= NULL;
	struct dlist_head	*entry	= NULL;

	entry = dlist_first( &sched->tasks );
	for(;;) {
		cur = dlist_entry( entry, struct coro_task, list );

		if( !cur->func )
			continue;

		/* direct call */
		switch( cur->state ) {
		case CORO_INIT:
			cur->state = CORO_RUNNING;
			sched->active_tasks++;

		case CORO_RUNNING:
#ifdef USE_SWAPCONTEXT
			if( swapcontext( &sched->env, &cur->env ) < 0 ) {
				/* something bad happened, kill coroutine */
				cur->state = CORO_SHUTDOWN;
			}
#else
			if( !sigsetjmp( sched->env, 0 ) )
				siglongjmp( cur->env, 0 );
#endif
			break;

                case CORO_SHUTDOWN:
			cur->state = CORO_FINISHED;
                        sched->active_tasks--;

		default:
			break;
		}

		if( !sched->active_tasks || sched->abort )
			break;

		entry = dlist_is_last( entry ) ? dlist_first( &sched->tasks ) : dlist_next( entry );
	}

	return 0;
}

int coro_pause( struct coro_task *tsk )
{
	if( tsk->state == CORO_RUNNING ) {
		tsk->state = CORO_PAUSED;
		return 0;
	}

	return -1;
}

int coro_continue( struct coro_task *tsk )
{
	if( tsk->state == CORO_PAUSED ) {
		tsk->state = CORO_RUNNING;
		return 0;
	}

	return -1;
}

struct coro_task * coro_find( const struct coro_task *tsk, const int id )
{
	struct dlist_head	*entry;
	struct coro_task	*p = NULL;

	if( id <= 0 || id > tsk->sched->last_id )
		return NULL;

	dlist_for_each( &tsk->sched->tasks, entry ) {
		p = dlist_entry( entry, struct coro_task, list );

		if( p->id == id )
			break;

		p = NULL;
	}

	return p;
}

inline int coro_yield( struct coro_task *tsk )
{
#ifdef USE_SWAPCONTEXT
	if( swapcontext( &tsk->env, &tsk->sched->env ) < 0 ) {
		/* something bad happened, kill coroutine */
		tsk->state = CORO_SHUTDOWN;
		return -1;
	}
#else
	if( !sigsetjmp( tsk->env, 0 ) ) {
		siglongjmp( tsk->sched->env, 0 );
	}
#endif
	return 0;
}

inline int coro_yieldto( struct coro_task *tsk, struct coro_task *to )
{
	if( !to || to->state != CORO_RUNNING )
		return -1;

#ifdef USE_SWAPCONTEXT
	if( swapcontext( &tsk->env, &to->env ) < 0 ) {
		/* something bad happened, kill coroutine */
		to->state = CORO_SHUTDOWN;
		return -1;
	}
#else
	if( !sigsetjmp( tsk->env, 0 ) ) {
		siglongjmp( to->env, 0 );
	}
#endif
	return 0;
}

inline void coro_return( struct coro_task *tsk )
{
	tsk->state = CORO_SHUTDOWN;
}
