#ifndef _DLIST_H
#define _DLIST_H

/*
 * Double-linked list implementation
 */
struct dlist_head;

/**
 * Double-linked list structure
 */
struct dlist_head {
	struct dlist_head *head;	/*!< pointer to list head */
	struct dlist_head *next;	/*!< next element (first) */
	struct dlist_head *prev;	/*!< previous element (last) */
};


/**
 * dlist_init_head
 * \brief	Initialize linked list
 * \param[in]	h	pointer to linked list head
 */
void dlist_init_head( struct dlist_head *h );

/**
 * dlist_entry
 * \brief Get current element
 */
#define dlist_entry(ptr, type, member) \
	container_of(ptr, type, member )



/**
 * dlist_insert_after
 * \brief	Insert new element after given one
 * \param[in]	elem	Element to insert after
 * \param[in]	new	New element
 */
void dlist_insert_after( struct dlist_head *elem, struct dlist_head *new );


/**
 * dlist_insert_head
 * \brief	Insert new element at first position
 * \param[in]	list	The list
 * \param[in]	new	New element
 */
#define dlist_insert_head(list, new) \
	dlist_insert_after( (list)->head, new )


/**
 * dlist_insert_before
 * \brief	Insert new element before given one
 * \param[in]	elem	Element to insert before
 * \param[in]	new	New element
 */
void dlist_insert_before( struct dlist_head *elem, struct dlist_head *new );


/**
 * dlist_insert_tail
 * \brief	Insert new element at the end of the list
 * \param[in]	list	The list
 * \param[in]	new	New element
 */
void dlist_insert_tail( struct dlist_head *list, struct dlist_head *new );


/**
 * dlist_remove
 * \brief	Remove element from the list
 * \param[in]	elem	Element to remove
 */
void dlist_remove( struct dlist_head *elem );


/**
 * dlist_first
 * \brief	Get first element in the list
 * \param[in]	elem	Any list element
 * \return	First element
 */
struct dlist_head * dlist_first( struct dlist_head *elem );


/**
 * dlist_last
 * \brief	Get last element in the list
 * \param[in]	elem	Any list element
 * \return	Last element
 */
struct dlist_head * dlist_last( struct dlist_head *elem );


/**
 * dlist_next
 * \brief	Get next element after given one
 * \param[in]	elem	Any list element
 * \return	Next element
 */
struct dlist_head * dlist_next( struct dlist_head *elem );


/**
 * dlist_prev
 * \brief	Get element before given one
 * \param[in]	elem	Any list element
 * \return	Previous element
 */
struct dlist_head * dlist_prev( struct dlist_head *elem );


/**
 * dlist_is_head
 * \brief	Check if the element passed is the list head
 * \param[in]	elem	Any list element
 * \return	0 if the element passed is not the head, 1 otherwise
 */
int dlist_is_head( struct dlist_head *elem );


/**
 * dlist_is_first
 * \brief	Check if the element passed is the first one
 * \param[in]	elem	Any list element
 * \return	0 if the element passed is not the first, 1 otherwise
 */
int dlist_is_first( struct dlist_head *elem );


/**
 * dlist_is_last
 * \brief	Check if the element passed is the last one
 * \param[in]	elem	Any list element
 * \return	0 if the element passed is not the last, 1 otherwise
 */
int dlist_is_last( struct dlist_head *elem );


/**
 * dlist_replace
 * \brief	Replace an element in the list with a new one
 * \param[in]	old	Element to replace
 * \param[in]	new	Replacement
 */
void dlist_replace( struct dlist_head *old, struct dlist_head *new );


/**
 * dlist_for_each
 * \brief	Walk the list of elements in a for like loop
 * \param[in]	head	The list
 * \param[in]	p	Pointer to store the current element
 * \note	Not dlist_remove() safe!
 */
#define dlist_for_each(head, p)		\
	for( p = (typeof(*p) *)(head)->next;	\
	p != (typeof(*p) *)(head);		\
	p = (typeof(*p) *)(p)->next )


/**
 * dlist_for_each_safe
 * \brief	Walk the list of elements in a for like loop
 * \param[in]	head	The list
 * \param[in]	n	Temporary storage	
 * \param[in]	p	Pointer to store the current element
 * \note	dlist_remove() safe version
 */
#define dlist_for_each_safe(head, n, p)	\
	for( p = (typeof(*p) *)(head)->next, n = (typeof(*p) *)(p)->next; \
	p != (typeof(*p) *)(head); 		\
	p = (typeof(*p) *)(n), n = (typeof(*p) *)(p)->next )


/**
 * dlist_for_each_reverse
 * \brief	Reverse-walk the list of elements in a for like loop
 * \param[in]	head	The list
 * \param[in]	p	Pointer to store the current element
 * \note	Not dlist_remove() safe!
 */
#define dlist_for_each_reverse(head, p)	\
	for( p = (typeof(*p) *)(head)->prev;	\
	p != (typeof(*p) *)(head);		\
	p = (typeof(*p) *)(p)->prev )


/**
 * dlist_is_empty
 * \brief	Check if the list has no elements
 * \param[in]	head	The list
 * \return	0 if the list is not empty, 1 otherwise
 */
#define dlist_is_empty(head) ({		\
	((head)->next == head);		\
	})


/**
 * dlist_num_entries
 * \brief	Get the number of elements in the list
 * \param[in]	head	The list
 * \return	number of entries
 */
#define dlist_num_entries(head) ({		\
	int _x = 0;				\
	struct dlist_head *p = (head);		\
	for( ; p && p->next && p->next != (head); _x++) p = p->next;	\
	_x; })


#endif

