#include <string.h>
#include <stdlib.h>
#include <stdint.h>
#include <errno.h>

#include "net.h"

/*
 * Private helper macros
 */
#define netaddr_v4(x)		((x)->addr.ip4)
#define netaddr_v6(x)		((x)->addr.ip6)
#define netaddr_family(x)	((x)->addr.sa.sa_family)


int netaddr_is_ipv4( const struct netaddr *na )
{
	return( netaddr_family(na) == AF_INET );
}

int netaddr_is_ipv6( const struct netaddr *na )
{
	return( netaddr_family(na) == AF_INET6 );
}

int netaddr_get_family( const struct netaddr *na )
{
	return netaddr_family(na);
}

int netaddr_set_family( struct netaddr *na, const int family )
{
	netaddr_family(na) = family;
	return 0;
}

struct netaddr *netaddr_create( const struct sockaddr *sa )
{
	struct netaddr *na;

	na = malloc( sizeof(struct netaddr) );
	if( !na ) {
		return NULL;
	}

	netaddr_init( na );

	if( !sa ) {
		return na;
	}

	if( netaddr_set(na, sa) < 0 ) {
		free( na );
		return NULL;
	}
	return na;
}

int netaddr_destroy( struct netaddr **na )
{
	if( !na || !*na ) {
		return -1;
	}

	free( *na );
	*na = NULL;

	return 0;
}

void netaddr_init( struct netaddr *na )
{
	memset( na, 0, sizeof(struct netaddr) );
}

unsigned short netaddr_get_port( const struct netaddr *na )
{
	if( !na ) {
		errno = EINVAL;
		return -1;
	}

	switch( netaddr_family(na) ) {
	case AF_INET:
		return netaddr_v4(na).sin_port;

	case AF_INET6:
		return netaddr_v6(na).sin6_port;
	}
	return 0;
}

int netaddr_set_port( struct netaddr *na, const unsigned short port )
{
	if( !na || !port ) {
		errno = EINVAL;
		return -1;
	}

	switch( netaddr_family(na) ) {
	case AF_INET:
		netaddr_v4(na).sin_port = port;
		break;

	case AF_INET6:
		netaddr_v6(na).sin6_port = port;
		break;

	default:
		return -1;
	}
	return 0;
}

int netaddr_set( struct netaddr *na, const struct sockaddr *sa )
{
	if( !na || !sa ) {
		errno = EINVAL;
		return -1;
	}

	switch( sa->sa_family ) {
	case AF_INET:
		netaddr_v4(na).sin_port = ((struct sockaddr_in *)sa)->sin_port;
		netaddr_v4(na).sin_family = AF_INET;
		memcpy( &netaddr_v4(na).sin_addr, &((struct sockaddr_in *)sa)->sin_addr, sizeof(struct in_addr) );
		break;

	case AF_INET6:
		netaddr_v6(na).sin6_port = ((struct sockaddr_in6 *)sa)->sin6_port;
		netaddr_v6(na).sin6_family = AF_INET6;
		memcpy( &netaddr_v6(na).sin6_addr, &((struct sockaddr_in6 *)sa)->sin6_addr, sizeof(struct in6_addr) );
		break;

	default:
		return -1;
	}
	return 0;
}

int netaddr_get( const struct netaddr *na, struct sockaddr *sa, socklen_t *salen )
{
	if( !na || !sa || !salen || !*salen ) {
		errno = EINVAL;
		return -1;
	}

	switch( netaddr_family(na) ) {
	case AF_INET:
		{
			struct sockaddr_in *ptr = (struct sockaddr_in *)sa;

			if( *salen < sizeof(struct sockaddr_in) ) {
				errno = ENOSPC;
				return -1;
			}

			ptr->sin_port = netaddr_v4(na).sin_port;
			ptr->sin_family = AF_INET;
			memcpy( &ptr->sin_addr, &netaddr_v4(na).sin_addr, sizeof(struct in_addr) );
			*salen = sizeof(struct sockaddr_in);
		}
		break;

	case AF_INET6:
		{
			struct sockaddr_in6 *ptr = (struct sockaddr_in6 *)sa;

			if( *salen < sizeof(struct sockaddr_in6) ) {
				errno = ENOSPC;
				return -1;
			}

			ptr->sin6_port = netaddr_v6(na).sin6_port;
			ptr->sin6_family = AF_INET6;
			memcpy( &ptr->sin6_addr, &netaddr_v6(na).sin6_addr, sizeof(struct in6_addr) );
			*salen = sizeof(struct sockaddr_in6);
		}
		break;

	default:
		return -1;
	}
	return 0;
}

const char * netaddr_ntop( const struct netaddr *na, char *buf, socklen_t buflen )
{
	if( !na || !buf || !buflen ) {
		errno = EINVAL;
		return NULL;
	}

	switch( netaddr_family(na) ) {
	case AF_INET:
		return inet_ntop( AF_INET, (void *)&netaddr_v4(na).sin_addr, buf, buflen );
		break;

	case AF_INET6:
		return inet_ntop( AF_INET6, (void *)&netaddr_v6(na).sin6_addr, buf, buflen );
		break;

	default:
		errno = EAFNOSUPPORT;
		return NULL;
	}
}

int netaddr_pton( struct netaddr *na, const char *buf )
{
	int	res = 0;
	void	*ptr;

	if( !na || !buf ) {
		errno = EINVAL;
		return -1;
	}

	switch( netaddr_family(na) ) {
	case AF_INET:
		ptr = (void *)&netaddr_v4(na).sin_addr;
		break;

	case AF_INET6:
		ptr = (void *)&netaddr_v6(na).sin6_addr;
		break;

	default:
		/*
		 * Try IPv4 first
		 */
		res = inet_pton( AF_INET, buf, (void *)&netaddr_v4(na).sin_addr );
		if( res <= 0 ) {
			/*
			 * Retry with IPv6
			 */
			res = inet_pton( AF_INET6, buf, (void *)&netaddr_v6(na).sin6_addr );
			if( res <= 0 ) {
				errno = EINVAL;
				return -1;
			}
			else {
				netaddr_family(na) = AF_INET6;
			}
		}
		else {
			netaddr_family(na) = AF_INET;
		}
		return 0;
	}

	/*
	 * address family already set
	 */
	res = inet_pton( netaddr_family(na), buf, ptr );
	if( res <= 0 ) {
		switch( errno ) {
		case 0:
			errno = EINVAL;
		}
		return -1;
	}
	return 0;
}
