/*
 * Julienne Walker's Red-Black tree implementation
 *
 * URL: http://eternallyconfuzzled.com/tuts/datastructures/jsw_tut_rbtree.aspx
 * License: Public-Domain
 */
#include <stdint.h>
#include <stdlib.h>
#include "rb_tree.h"

enum {
	RB_DIR_LEFT = 0,
	RB_DIR_RIGHT
};

enum {
	RB_COLOR_BLACK = 0,
	RB_COLOR_RED
};

struct rb_node {
	uint32_t key;
	uint32_t color;
	void *value;
	struct rb_node *link[2];
};

#define rb_is_red(p)	(p && p->color == RB_COLOR_RED)


static struct rb_node *rb_node_new( uint32_t key, void *value )
{
	struct rb_node *rn = malloc( sizeof(struct rb_node) );
	if( rn ) {
		rn->key   = key;
		rn->value = value;
		rn->color = RB_COLOR_RED;
		rn->link[RB_DIR_LEFT]  = NULL;
		rn->link[RB_DIR_RIGHT] = NULL;
	}

	return rn;
}

static struct rb_node *rb_rotate_single( struct rb_node *root, int dir )
{
	struct rb_node *save = root->link[!dir];

	root->link[!dir] = save->link[dir];
	save->link[dir]  = root;

	root->color = RB_COLOR_RED;
	save->color = RB_COLOR_BLACK;

	return save;
}

static struct rb_node *rb_rotate_double( struct rb_node *root, int dir )
{
	root->link[!dir] = rb_rotate_single( root->link[!dir], !dir );
	return rb_rotate_single( root, dir );
}



/*
 * Public API
 */

void *rb_find( struct rb_tree *tree, uint32_t key )
{
	struct rb_node *p;
	int dir;

	if( !tree || !tree->root )
		return NULL;

	p = tree->root;

	while( p ) {
		if( p->key == key ) {
			return p->value;
		}

		dir = p->key < key;
		p   = p->link[dir];
	}

	return NULL;
}

int rb_insert( struct rb_tree *tree, uint32_t key, void *value )
{
	if( !tree->root ) {
		/* Empty tree case */
		tree->root = rb_node_new( key, value );
		if ( !tree->root )
			return -1;
	}
	else {
		struct rb_node head = { 0 }; /* False tree root */
		struct rb_node *g, *t;     /* Grandparent & parent */
		struct rb_node *p, *q;     /* Iterator & parent */
		int dir = RB_DIR_LEFT, last;

		/* Set up helpers */
		t = &head;
		g = p = NULL;
		q = t->link[RB_DIR_RIGHT] = tree->root;

		/* Search down the tree */
		while(1) {
			if( !q ) {
				/* Insert new node at the bottom */
				p->link[dir] = q = rb_node_new( key, value );
				if( !q )
					return -1;
			}
			else if( rb_is_red(q->link[RB_DIR_LEFT]) && rb_is_red(q->link[RB_DIR_RIGHT]) ) {
				/* Color flip */
				q->color = RB_COLOR_RED;
				q->link[RB_DIR_LEFT]->color  = RB_COLOR_BLACK;
				q->link[RB_DIR_RIGHT]->color = RB_COLOR_BLACK;
			}

			/* Fix red violation */
			if( rb_is_red(q) && rb_is_red(p) ) {
				int dir2 = t->link[RB_DIR_RIGHT] == g;

				if( q == p->link[last] ) {
					t->link[dir2] = rb_rotate_single( g, !last );
				}
				else {
					t->link[dir2] = rb_rotate_double( g, !last );
				}
			}

			/* Stop if found */
			if( q->key == key )
				break;

			last = dir;
			dir  = q->key < key;

			/* Update helpers */
			if( g )
				t = g;
			g = p;
			p = q;
			q = q->link[dir];
		}

		/* Update root */
		tree->root = head.link[RB_DIR_RIGHT];
	}

	/* Make root black */
	tree->root->color = RB_COLOR_BLACK;

	return 0;
}


int rb_remove( struct rb_tree *tree, uint32_t key )
{
	struct rb_node head = { 0 };
	struct rb_node *q, *p, *g;
	struct rb_node *f = NULL;
	int dir = RB_DIR_RIGHT;

	if( !tree || !tree->root )
		return -1;

	/* Set up helpers */
	q = &head;
	g = p = NULL;
	q->link[dir] = tree->root;

	/* Search and push a red down */
	while( q->link[dir] ) {
		int last = dir;

		/* Update helpers */
		g = p;
		p = q;
		q = q->link[dir];
		dir = q->key < key;

		/* Save found node */
		if( q->key == key ) {
			f = q;
		}

		/* Push the red node down */
		if( !rb_is_red(q) && !is_red(q->link[dir]) ) {

			if( rb_is_red(q->link[!dir]) ) {
				p = p->link[last] = rb_rotate_single( q, dir );
			}
			else if( !rb_is_red(q->link[!dir]) ) {
				struct rb_node *s = p->link[!last];

				if( s ) {
					if( !rb_is_red(s->link[!last]) && !rb_is_red(s->link[last]) ) {
						/* Color flip */
						p->color = RB_COLOR_BLACK;
						s->color = RB_COLOR_RED;
						q->color = RB_COLOR_RED;
					}
					else {
						int dir2 = g->link[RB_DIR_RIGHT] == p;

						if( rb_is_red(s->link[last]) ) {
							g->link[dir2] = rb_rotate_double( p, last );
						}
						else if( rb_is_red(s->link[!last]) ) {
							g->link[dir2] = rb_rotate_single( p, last );
						}

						/* Ensure correct coloring */
						q->color = g->link[dir2]->color = RB_COLOR_RED;
						g->link[dir2]->link[RB_DIR_LEFT]->color  = RB_COLOR_BLACK;
						g->link[dir2]->link[RB_DIR_RIGHT]->color = RB_COLOR_BLACK;
					}
				}
			}
		}
	}

	/* Replace and remove if found */
	if( f ) {
		f->key = q->key;
		p->link[p->link[RB_DIR_RIGHT] == q] = q->link[q->link[RB_DIR_LEFT] == NULL];
		free ( q );
	}

	/* Update root and make it black */
	tree->root = head.link[RB_DIR_RIGHT];
	if( tree->root ) {
		tree->root->color = RB_COLOR_BLACK;
	}

	return 0;
}
